@@ -1,10 +1,10 @@
 #include "system.h"
 
+#include <sys/types.h>
+#include <sys/wait.h>
 #include <errno.h>
 #include <unistd.h>
 
-#define _RPMSQ_INTERNAL
-#include <rpm/rpmsq.h>
 #include <rpm/rpmfileutil.h>
 #include <rpm/rpmmacro.h>
 #include <rpm/rpmio.h>
@@ -16,6 +16,14 @@
 
 #include "debug.h"
 
+struct rpmScript_s {
+    rpmTagVal tag;		/* script tag */
+    char **args;		/* scriptlet call arguments */
+    char *body;			/* script body */
+    char *descr;		/* description for logging */
+    rpmscriptFlags flags;	/* flags to control operation */
+};
+
 /**
  * Run internal Lua script.
  */
@@ -27,14 +35,13 @@ static rpmRC runLuaScript(int selinux, ARGV_const_t prefixes,
 #ifdef WITH_LUA
     ARGV_t argv = argvp ? *argvp : NULL;
     rpmlua lua = NULL; /* Global state. */
-    rpmluav var;
+    rpmluav var = rpmluavNew();
     int cwd = -1;
 
     rpmlog(RPMLOG_DEBUG, "%s: running <lua> scriptlet.\n", sname);
 
     /* Create arg variable */
     rpmluaPushTable(lua, "arg");
-    var = rpmluavNew();
     rpmluavSetListMode(var, 1);
     if (argv) {
 	char **p;
@@ -51,27 +58,29 @@ static rpmRC runLuaScript(int selinux, ARGV_const_t prefixes,
 	rpmluavSetValueNum(var, arg2);
 	rpmluaSetVar(lua, var);
     }
-    var = rpmluavFree(var);
     rpmluaPop(lua);
 
     /* Lua scripts can change our cwd and umask, save and restore */
     /* XXX TODO: use cwd from chroot state to save unnecessary open here */
     cwd = open(".", O_RDONLY);
     if (cwd != -1) {
-	int xx;
 	mode_t oldmask = umask(0);
 	umask(oldmask);
 
 	if (chdir("/") == 0 && rpmluaRunScript(lua, script, sname) == 0) {
 	    rc = RPMRC_OK;
 	}
-	/* XXX no way to return error from restore meaningfully atm */
-	xx = fchdir(cwd);
+	/* This failing would be fatal, return something different for it... */
+	if (fchdir(cwd)) {
+	    rpmlog(RPMLOG_ERR, _("Unable to restore current directory: %m"));
+	    rc = RPMRC_NOTFOUND;
+	}
 	close(cwd);
 	umask(oldmask);
     }
 
     rpmluaDelVar(lua, "arg");
+    rpmluavFree(var);
 
 #else
     rpmlog(lvl, _("<lua> scriptlet support not built in\n"));
@@ -133,7 +142,7 @@ static void doScriptExec(int selinux, ARGV_const_t argv, ARGV_const_t prefixes,
 	    path = ipath;
 
 	xx = setenv("PATH", path, 1);
-	ipath = _free(ipath);
+	free(ipath);
     }
 
     for (ARGV_const_t pf = prefixes; pf && *pf; pf++) {
@@ -166,6 +175,30 @@ static void doScriptExec(int selinux, ARGV_const_t argv, ARGV_const_t prefixes,
     _exit(127); /* exit 127 for compatibility with bash(1) */
 }
 
+static char * writeScript(const char *cmd, const char *script)
+{
+    char *fn = NULL;
+    size_t slen = strlen(script);
+    int ok = 0;
+    FD_t fd = rpmMkTempFile("/", &fn);
+
+    if (Ferror(fd))
+	goto exit;
+
+    if (rpmIsDebug() && (rstreq(cmd, "/bin/sh") || rstreq(cmd, "/bin/bash"))) {
+	static const char set_x[] = "set -x\n";
+	/* Assume failures will be caught by the write below */
+	Fwrite(set_x, sizeof(set_x[0]), sizeof(set_x)-1, fd);
+    }
+
+    ok = (Fwrite(script, sizeof(script[0]), slen, fd) == slen);
+
+exit:
+    if (!ok) fn = _free(fn);
+    Fclose(fd);
+    return fn;
+}
+
 /**
  * Run an external script.
  */
@@ -175,33 +208,21 @@ static rpmRC runExtScript(int selinux, ARGV_const_t prefixes,
 {
     FD_t out = NULL;
     char * fn = NULL;
-    int xx;
+    pid_t pid, reaped;
+    int status;
     rpmRC rc = RPMRC_FAIL;
-    struct rpmsqElem sq;
-
-    memset(&sq, 0, sizeof(sq));
-    sq.reaper = 1;
 
     rpmlog(RPMLOG_DEBUG, "%s: scriptlet start\n", sname);
 
     if (script) {
-	FD_t fd = rpmMkTempFile("/", &fn);
-	if (fd == NULL || Ferror(fd)) {
-	    rpmlog(RPMLOG_ERR, _("Couldn't create temporary file for %s: %s\n"),
+	fn = writeScript(*argvp[0], script);
+	if (fn == NULL) {
+	    rpmlog(RPMLOG_ERR,
+		   _("Couldn't create temporary file for %s: %s\n"),
 		   sname, strerror(errno));
 	    goto exit;
 	}
 
-	if (rpmIsDebug() &&
-	    (rstreq(*argvp[0], "/bin/sh") || rstreq(*argvp[0], "/bin/bash")))
-	{
-	    static const char set_x[] = "set -x\n";
-	    xx = Fwrite(set_x, sizeof(set_x[0]), sizeof(set_x)-1, fd);
-	}
-
-	xx = Fwrite(script, sizeof(script[0]), strlen(script), fd);
-	xx = Fclose(fd);
-
 	argvAdd(argvp, fn);
 	if (arg1 >= 0) {
 	    argvAddNum(argvp, arg1);
@@ -229,33 +250,34 @@ static rpmRC runExtScript(int selinux, ARGV_const_t prefixes,
 	goto exit;
     }
 
-    xx = rpmsqFork(&sq);
-    if (sq.child == 0) {
+    pid = fork();
+    if (pid == (pid_t) -1) {
+	rpmlog(RPMLOG_ERR, _("Couldn't fork %s: %s\n"),
+		sname, strerror(errno));
+	goto exit;
+    } else if (pid == 0) {/* Child */
 	rpmlog(RPMLOG_DEBUG, "%s: execv(%s) pid %d\n",
 	       sname, *argvp[0], (unsigned)getpid());
 	doScriptExec(selinux, *argvp, prefixes, scriptFd, out);
     }
 
-    if (sq.child == (pid_t)-1) {
-	rpmlog(RPMLOG_ERR, _("Couldn't fork %s: %s\n"), sname, strerror(errno));
-	goto exit;
-    }
-
-    rpmsqWait(&sq);
+    do {
+	reaped = waitpid(pid, &status, 0);
+    } while (reaped == -1 && errno == EINTR);
 
     rpmlog(RPMLOG_DEBUG, "%s: waitpid(%d) rc %d status %x\n",
-	   sname, (unsigned)sq.child, (unsigned)sq.reaped, sq.status);
+	   sname, (unsigned)pid, (unsigned)reaped, status);
 
-    if (sq.reaped < 0) {
+    if (reaped < 0) {
 	rpmlog(lvl, _("%s scriptlet failed, waitpid(%d) rc %d: %s\n"),
-		 sname, sq.child, sq.reaped, strerror(errno));
-    } else if (!WIFEXITED(sq.status) || WEXITSTATUS(sq.status)) {
-      	if (WIFSIGNALED(sq.status)) {
+		 sname, pid, reaped, strerror(errno));
+    } else if (!WIFEXITED(status) || WEXITSTATUS(status)) {
+      	if (WIFSIGNALED(status)) {
 	    rpmlog(lvl, _("%s scriptlet failed, signal %d\n"),
-                   sname, WTERMSIG(sq.status));
+                   sname, WTERMSIG(status));
 	} else {
 	    rpmlog(lvl, _("%s scriptlet failed, exit status %d\n"),
-		   sname, WEXITSTATUS(sq.status));
+		   sname, WEXITSTATUS(status));
 	}
     } else {
 	/* if we get this far we're clear */
@@ -264,12 +286,12 @@ static rpmRC runExtScript(int selinux, ARGV_const_t prefixes,
 
 exit:
     if (out)
-	xx = Fclose(out);	/* XXX dup'd STDOUT_FILENO */
+	Fclose(out);	/* XXX dup'd STDOUT_FILENO */
 
-    if (script) {
+    if (fn) {
 	if (!rpmIsDebug())
-	    xx = unlink(fn);
-	fn = _free(fn);
+	    unlink(fn);
+	free(fn);
     }
     return rc;
 }
@@ -350,6 +372,66 @@ static const char * tag2sln(rpmTagVal tag)
     return "%unknownscript";
 }
 
+static rpmScript rpmScriptNew(Header h, rpmTagVal tag, const char *body,
+			      rpmscriptFlags flags)
+{
+    char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
+    rpmScript script = xcalloc(1, sizeof(*script));
+    script->tag = tag;
+    script->flags = flags;
+    script->body = (body != NULL) ? xstrdup(body) : NULL;
+    rasprintf(&script->descr, "%s(%s)", tag2sln(tag), nevra);
+
+    /* macros need to be expanded before possible queryformat */
+    if (script->body && (script->flags & RPMSCRIPT_FLAG_EXPAND)) {
+	char *body = rpmExpand(script->body, NULL);
+	free(script->body);
+	script->body = body;
+    }
+    if (script->body && (script->flags & RPMSCRIPT_FLAG_QFORMAT)) {
+	/* XXX TODO: handle queryformat errors */
+	char *body = headerFormat(h, script->body, NULL);
+	free(script->body);
+	script->body = body;
+    }
+
+    free(nevra);
+    return script;
+}
+
+rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag, uint32_t ix)
+{
+    rpmScript script = NULL;
+    struct rpmtd_s tscripts, tprogs, tflags;
+    headerGetFlags hgflags = HEADERGET_MINMEM;
+
+    headerGet(h, RPMTAG_TRIGGERSCRIPTS, &tscripts, hgflags);
+    headerGet(h, RPMTAG_TRIGGERSCRIPTPROG, &tprogs, hgflags);
+    headerGet(h, RPMTAG_TRIGGERSCRIPTFLAGS, &tflags, hgflags);
+    
+    if (rpmtdSetIndex(&tscripts, ix) >= 0 && rpmtdSetIndex(&tprogs, ix) >= 0) {
+	rpmscriptFlags sflags = 0;
+	const char *prog = rpmtdGetString(&tprogs);
+
+	if (rpmtdSetIndex(&tflags, ix) >= 0)
+	    sflags = rpmtdGetNumber(&tflags);
+
+	script = rpmScriptNew(h, triggerTag, rpmtdGetString(&tscripts), sflags);
+
+	/* hack up a hge-style NULL-terminated array */
+	script->args = xmalloc(2 * sizeof(*script->args) + strlen(prog) + 1);
+	script->args[0] = (char *)(script->args + 2);
+	script->args[1] = NULL;
+	strcpy(script->args[0], prog);
+    }
+
+    rpmtdFreeData(&tscripts);
+    rpmtdFreeData(&tprogs);
+    rpmtdFreeData(&tflags);
+
+    return script;
+}
+
 rpmScript rpmScriptFromTag(Header h, rpmTagVal scriptTag)
 {
     rpmScript script = NULL;
@@ -357,31 +439,14 @@ rpmScript rpmScriptFromTag(Header h, rpmTagVal scriptTag)
 
     if (headerIsEntry(h, scriptTag) || headerIsEntry(h, progTag)) {
 	struct rpmtd_s prog;
-	char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
-	rpmscriptFlags flags = headerGetNumber(h, getFlagTag(scriptTag));
 
-	script = xcalloc(1, sizeof(*script));
-	script->tag = scriptTag;
-	rasprintf(&script->descr, "%s(%s)", tag2sln(scriptTag), nevra);
-	script->body = headerGetAsString(h, scriptTag);
-
-	/* macros need to be expanded before possible queryformat */
-	if (script->body && (flags & RPMSCRIPT_EXPAND)) {
-	    char *body = rpmExpand(script->body, NULL);
-	    free(script->body);
-	    script->body = body;
-	}
-	if (script->body && (flags & RPMSCRIPT_QFORMAT)) {
-	    /* XXX TODO: handle queryformat errors */
-	    char *body = headerFormat(h, script->body, NULL);
-	    free(script->body);
-	    script->body = body;
-	}
+	script = rpmScriptNew(h, scriptTag,
+			      headerGetString(h, scriptTag),
+			      headerGetNumber(h, getFlagTag(scriptTag)));
 
 	if (headerGet(h, progTag, &prog, (HEADERGET_ALLOC|HEADERGET_ARGV))) {
 	    script->args = prog.data;
 	}
-	free(nevra);
     }
     return script;
 }
@@ -396,3 +461,8 @@ rpmScript rpmScriptFree(rpmScript script)
     }
     return NULL;
 }
+
+rpmTagVal rpmScriptTag(rpmScript script)
+{
+    return (script != NULL) ? script->tag : RPMTAG_NOT_FOUND;
+}
