@@ -227,46 +227,12 @@ static void rpmtsUpdateDSI(const rpmts ts, dev_t dev, const char *dirName,
     if (dsi->ineeded < dsi->oineeded) dsi->oineeded = dsi->ineeded;
 }
 
-/* return DSI of the device the rpmdb lives on */
-static rpmDiskSpaceInfo rpmtsDbDSI(const rpmts ts) {
-    const char *dbhome = rpmdbHome(rpmtsGetRdb(ts));
-    struct stat sb;
-    int rc;
-
-    rc = stat(dbhome, &sb);
-    if (rc) {
-	return NULL;
-    }
-    return rpmtsGetDSI(ts, sb.st_dev, dbhome);
-}
-
-/* Update DSI for changing size of the rpmdb */
-static void rpmtsUpdateDSIrpmDBSize(const rpmte p,
-				    rpmDiskSpaceInfo dsi) {
-    rpm_loff_t headerSize;
-    int64_t bneeded;
-
-    /* XXX somehow we can end up here with bsize 0 (RhBug:671056) */
-    if (dsi == NULL || dsi->bsize == 0) return;
-
-    headerSize = rpmteHeaderSize(p);
-    bneeded = BLOCK_ROUND(headerSize, dsi->bsize);
-    /* REMOVE doesn't neccessarily shrink the database */
-    if (rpmteType(p) == TR_ADDED) {
-	/* guessing that db grows 4 times more than the header size */
-	dsi->bneeded += (bneeded * 4);
-    }
-}
-
-
 static void rpmtsCheckDSIProblems(const rpmts ts, const rpmte te)
 {
     rpmDiskSpaceInfo dsi = ts->dsi;
 
     if (dsi == NULL || !dsi->bsize)
 	return;
-    if (rpmfiFC(rpmteFI(te)) <= 0)
-	return;
 
     for (; dsi->bsize; dsi++) {
 
@@ -288,7 +254,8 @@ static void rpmtsCheckDSIProblems(const rpmts ts, const rpmte te)
     }
 }
 
-static void rpmtsFreeDSI(rpmts ts){
+static void rpmtsFreeDSI(rpmts ts)
+{
     rpmDiskSpaceInfo dsi;
     if (ts == NULL)
 	return;
@@ -310,7 +277,7 @@ static uint64_t countFiles(rpmts ts)
     rpmte p;
     while ((p = rpmtsiNext(pi, 0)) != NULL)
 	fc += rpmfiFC(rpmteFI(p));
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
     return fc;
 }
 
@@ -324,61 +291,62 @@ static uint64_t countFiles(rpmts ts)
  * @param reportConflicts
  */
 /* XXX only ts->{probs,rpmdb} modified */
-static int handleInstInstalledFile(const rpmts ts, rpmte p, rpmfi fi,
-				   Header otherHeader, rpmfi otherFi,
+static void handleInstInstalledFile(const rpmts ts, rpmte p, rpmfi fi, int fx,
+				   Header otherHeader, rpmfi otherFi, int ofx,
 				   int beingRemoved)
 {
-    unsigned int fx = rpmfiFX(fi);
     rpmfs fs = rpmteGetFileStates(p);
-    int isCfgFile = ((rpmfiFFlags(otherFi) | rpmfiFFlags(fi)) & RPMFILE_CONFIG);
+    int isCfgFile = ((rpmfiFFlagsIndex(otherFi, ofx) | rpmfiFFlagsIndex(fi, fx)) & RPMFILE_CONFIG);
 
     if (XFA_SKIPPING(rpmfsGetAction(fs, fx)))
-	return 0;
+	return;
 
-    if (rpmfiCompare(otherFi, fi)) {
+    if (rpmfiCompareIndex(otherFi, ofx, fi, fx)) {
 	rpm_color_t tscolor = rpmtsColor(ts);
 	rpm_color_t prefcolor = rpmtsPrefColor(ts);
-	rpm_color_t FColor = rpmfiFColor(fi) & tscolor;
-	rpm_color_t oFColor = rpmfiFColor(otherFi) & tscolor;
+	rpm_color_t FColor = rpmfiFColorIndex(fi, fx) & tscolor;
+	rpm_color_t oFColor = rpmfiFColorIndex(otherFi, ofx) & tscolor;
 	int rConflicts;
+	char rState = RPMFILE_STATE_REPLACED;
 
 	rConflicts = !(beingRemoved || (rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACEOLDFILES));
 	/* Resolve file conflicts to prefer Elf64 (if not forced). */
-	if (tscolor != 0 && FColor != 0 && FColor != oFColor) {
+	if (tscolor != 0 && FColor != 0 && oFColor != 0 && FColor != oFColor) {
 	    if (oFColor & prefcolor) {
 		rpmfsSetAction(fs, fx, FA_SKIPCOLOR);
 		rConflicts = 0;
 	    } else if (FColor & prefcolor) {
 		rpmfsSetAction(fs, fx, FA_CREATE);
 		rConflicts = 0;
+		rState = RPMFILE_STATE_WRONGCOLOR;
 	    }
 	}
 
 	if (rConflicts) {
 	    char *altNEVR = headerGetAsString(otherHeader, RPMTAG_NEVRA);
-	    rpmteAddProblem(p, RPMPROB_FILE_CONFLICT, altNEVR, rpmfiFN(fi),
+	    char *fn = rpmfiFNIndex(fi, fx);
+	    rpmteAddProblem(p, RPMPROB_FILE_CONFLICT, altNEVR, fn,
 			    headerGetInstance(otherHeader));
+	    free(fn);
 	    free(altNEVR);
 	}
 
 	/* Save file identifier to mark as state REPLACED. */
 	if ( !(isCfgFile || XFA_SKIPPING(rpmfsGetAction(fs, fx))) ) {
 	    if (!beingRemoved)
-		rpmfsAddReplaced(rpmteGetFileStates(p), rpmfiFX(fi),
-				 headerGetInstance(otherHeader),
-				 rpmfiFX(otherFi));
+		rpmfsAddReplaced(rpmteGetFileStates(p), fx, rState,
+				 headerGetInstance(otherHeader), ofx);
 	}
     }
 
     /* Determine config file dispostion, skipping missing files (if any). */
     if (isCfgFile) {
 	int skipMissing = ((rpmtsFlags(ts) & RPMTRANS_FLAG_ALLFILES) ? 0 : 1);
-	rpmFileAction action = rpmfiDecideFate(otherFi, fi, skipMissing);
+	rpmFileAction action;
+	action = rpmfiDecideFateIndex(otherFi, ofx, fi, fx, skipMissing);
 	rpmfsSetAction(fs, fx, action);
     }
-    rpmfiSetFReplacedSize(fi, rpmfiFSize(otherFi));
-
-    return 0;
+    rpmfiSetFReplacedSizeIndex(fi, fx, rpmfiFSizeIndex(otherFi, ofx));
 }
 
 /**
@@ -388,15 +356,14 @@ static int handleInstInstalledFile(const rpmts ts, rpmte p, rpmfi fi,
 static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 {
     rpm_loff_t fixupSize = 0;
-    const char * fn;
     int i, j;
     rpm_color_t tscolor = rpmtsColor(ts);
     rpm_color_t prefcolor = rpmtsPrefColor(ts);
     rpmfs fs = rpmteGetFileStates(p);
     rpmfs otherFs;
+    rpm_count_t fc = rpmfiFC(fi);
 
-    fi = rpmfiInit(fi, 0);
-    while ((i = rpmfiNext(fi)) >= 0) {
+    for (i = 0; i < fc; i++) {
 	rpm_color_t oFColor, FColor;
 	struct fingerPrint_s * fiFps;
 	int otherPkgNum, otherFileNum;
@@ -410,11 +377,10 @@ static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 	if (XFA_SKIPPING(rpmfsGetAction(fs, i)))
 	    continue;
 
-	fn = rpmfiFN(fi);
 	fiFps = rpmfiFpsIndex(fi, i);
-	FFlags = rpmfiFFlags(fi);
-	FMode = rpmfiFMode(fi);
-	FColor = rpmfiFColor(fi);
+	FFlags = rpmfiFFlagsIndex(fi, i);
+	FMode = rpmfiFModeIndex(fi, i);
+	FColor = rpmfiFColorIndex(fi, i);
 	FColor &= tscolor;
 
 	fixupSize = 0;
@@ -448,8 +414,13 @@ static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 	 *
 	 */
 
-	/* Locate this overlapped file in the set of added/removed packages. */
-	for (j = 0; j < numRecs && recs[j].p != p; j++)
+	/*
+	 * Locate this overlapped file in the set of added/removed packages,
+	 * including the package owning it: a package can have self-conflicting
+	 * files when directory symlinks are present. Don't compare a file
+	 * with itself though...
+	 */
+	for (j = 0; j < numRecs && !(recs[j].p == p && recs[j].fileno == i); j++)
 	    {};
 
 	/* Find what the previous disposition of this file was. */
@@ -468,14 +439,12 @@ static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 	    if (rpmteType(p) == TR_ADDED && rpmteType(otherTe) != TR_ADDED)
 		continue;
 
-	    (void) rpmfiSetFX(otherFi, otherFileNum);
-
 	    /* XXX Happens iff fingerprint for incomplete package install. */
 	    if (rpmfsGetAction(otherFs, otherFileNum) != FA_UNKNOWN)
 		break;
 	}
 
-	oFColor = rpmfiFColor(otherFi);
+	oFColor = rpmfiFColorIndex(otherFi, otherFileNum);
 	oFColor &= tscolor;
 
 	switch (rpmteType(p)) {
@@ -490,7 +459,7 @@ static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 		rpmFileAction action;
 		if (rpmfsGetAction(fs, i) != FA_UNKNOWN)
 		    break;
-		if (rpmfiConfigConflict(fi)) {
+		if (rpmfiConfigConflictIndex(fi, i)) {
 		    /* Here is a non-overlapped pre-existing config file. */
 		    action = (FFlags & RPMFILE_NOREPLACE) ?
 			      FA_ALTNAME : FA_BACKUP;
@@ -503,12 +472,12 @@ static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 
 assert(otherFi != NULL);
 	    /* Mark added overlapped non-identical files as a conflict. */
-	    if (rpmfiCompare(otherFi, fi)) {
+	    if (rpmfiCompareIndex(otherFi, otherFileNum, fi, i)) {
 		int rConflicts;
 
 		rConflicts = reportConflicts;
 		/* Resolve file conflicts to prefer Elf64 (if not forced) ... */
-		if (tscolor != 0) {
+		if (tscolor != 0 && FColor != 0 && oFColor != 0 && FColor != oFColor) {
 		    if (FColor & prefcolor) {
 			/* ... last file of preferred colour is installed ... */
 			if (!XFA_SKIPPING(rpmfsGetAction(fs, i)))
@@ -526,15 +495,23 @@ assert(otherFi != NULL);
 		    done = 1;
 		}
 		if (rConflicts) {
+		    char *fn = rpmfiFNIndex(fi, i);
 		    rpmteAddProblem(p, RPMPROB_NEW_FILE_CONFLICT,
 				    rpmteNEVRA(otherTe), fn, 0);
+		    free(fn);
+		}
+	    } else {
+		/* Skip create on all but the first instance of a shared file */
+		if (rpmfsGetAction(otherFs, otherFileNum) != FA_UNKNOWN) {
+		    rpmfsSetAction(fs, i, FA_SKIP);
+		    done = 1;
 		}
 	    }
 
 	    /* Try to get the disk accounting correct even if a conflict. */
-	    fixupSize = rpmfiFSize(otherFi);
+	    fixupSize = rpmfiFSizeIndex(otherFi, otherFileNum);
 
-	    if (rpmfiConfigConflict(fi)) {
+	    if (rpmfiConfigConflictIndex(fi, i)) {
 		/* Here is an overlapped  pre-existing config file. */
 		rpmFileAction action;
 		action = (FFlags & RPMFILE_NOREPLACE) ? FA_ALTNAME : FA_SKIP;
@@ -559,33 +536,35 @@ assert(otherFi != NULL);
 	    }
 	    if (XFA_SKIPPING(rpmfsGetAction(fs, i)))
 		break;
-	    if (rpmfiFState(fi) != RPMFILE_STATE_NORMAL)
+	    if (rpmfiFStateIndex(fi, i) != RPMFILE_STATE_NORMAL)
 		break;
-	    if (!(S_ISREG(FMode) && (FFlags & RPMFILE_CONFIG))) {
-		rpmfsSetAction(fs, i, FA_ERASE);
-		break;
-	    }
 		
-	    /* Here is a pre-existing modified config file that needs saving. */
-	    {	int algo = 0;
+	    /* Pre-existing modified config files need to be saved. */
+	    if (S_ISREG(FMode) && (FFlags & RPMFILE_CONFIG)) {
+	    	int algo = 0;
 		size_t diglen = 0;
 		const unsigned char *digest;
-		if ((digest = rpmfiFDigest(fi, &algo, &diglen))) {
+		if ((digest = rpmfiFDigestIndex(fi, i, &algo, &diglen))) {
 		    unsigned char fdigest[diglen];
-		    if (!rpmDoDigest(algo, fn, 0, fdigest, NULL) &&
-			memcmp(digest, fdigest, diglen)) {
+		    char *fn = rpmfiFNIndex(fi, i);
+		    int modified = (!rpmDoDigest(algo, fn, 0, fdigest, NULL) &&
+				    memcmp(digest, fdigest, diglen));
+		    free(fn);
+		    if (modified) {
 			rpmfsSetAction(fs, i, FA_BACKUP);
 			break;
 		    }
 		}
 	    }
+	
+	    /* Otherwise, we can just erase. */
 	    rpmfsSetAction(fs, i, FA_ERASE);
 	    break;
 	}
 
 	/* Update disk space info for a file. */
 	rpmtsUpdateDSI(ts, fiFps->entry->dev, fiFps->entry->dirName,
-		       rpmfiFSize(fi), rpmfiFReplacedSize(fi),
+		       rpmfiFSizeIndex(fi, i), rpmfiFReplacedSizeIndex(fi, i),
 		       fixupSize, rpmfsGetAction(fs, i));
 
     }
@@ -859,12 +838,11 @@ rpmdbMatchIterator rpmFindBaseNamesInDB(rpmts ts, uint64_t fileCount)
     rpmtsi pi;  rpmte p;
     rpmfi fi;
     rpmdbMatchIterator mi;
-    int xx;
     int oc = 0;
     const char * baseName;
 
     rpmStringSet baseNames = rpmStringSetCreate(fileCount, 
-					hashFunctionString, strcmp, NULL);
+					rstrhash, strcmp, NULL);
 
     mi = rpmdbNewIterator(rpmtsGetRdb(ts), RPMDBI_BASENAMES);
 
@@ -885,11 +863,11 @@ rpmdbMatchIterator rpmFindBaseNamesInDB(rpmts ts, uint64_t fileCount)
 	    keylen = strlen(baseName);
 	    if (keylen == 0)
 		keylen++;	/* XXX "/" fixup. */
-	    xx = rpmdbExtendIterator(mi, baseName, keylen);
+	    rpmdbExtendIterator(mi, baseName, keylen);
 	    rpmStringSetAddEntry(baseNames, baseName);
 	 }
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
     rpmStringSetFree(baseNames);
 
     rpmdbSortIterator(mi);
@@ -912,7 +890,6 @@ void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrint
     rpmfs fs;
     rpmfi otherFi=NULL;
     int j;
-    int xx;
     unsigned int fileNum;
     const char * oldDir;
 
@@ -990,15 +967,14 @@ void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrint
 		switch (rpmteType(p)) {
 		case TR_ADDED:
 		    if (!otherFi) {
+			/* XXX What to do if this fails? */
 		        otherFi = rpmfiNew(ts, h, RPMTAG_BASENAMES, RPMFI_KEEPHEADER);
 		    }
-		    rpmfiSetFX(fi, recs[j].fileno);
-		    rpmfiSetFX(otherFi, fileNum);
-		    xx = handleInstInstalledFile(ts, p, fi, h, otherFi, beingRemoved);
+		    handleInstInstalledFile(ts, p, fi, recs[j].fileno,
+					    h, otherFi, fileNum, beingRemoved);
 		    break;
 		case TR_REMOVED:
 		    if (!beingRemoved) {
-		        rpmfiSetFX(fi, recs[j].fileno);
 			if (*rpmtdGetChar(&ostates) == RPMFILE_STATE_NORMAL)
 			    rpmfsSetAction(fs, recs[j].fileno, FA_SKIP);
 		    }
@@ -1019,7 +995,7 @@ void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrint
 	h = newheader;
     }
 
-    mi = rpmdbFreeIterator(mi);
+    rpmdbFreeIterator(mi);
 }
 
 #define badArch(_a) (rpmMachineScore(RPM_MACHTABLE_INSTARCH, (_a)) == 0)
@@ -1041,7 +1017,6 @@ static rpmps checkProblems(rpmts ts)
     /* XXX Only added packages need be checked. */
     rpmlog(RPMLOG_DEBUG, "sanity checking %d elements\n", rpmtsNElements(ts));
     while ((p = rpmtsiNext(pi, TR_ADDED)) != NULL) {
-	rpmdbMatchIterator mi;
 
 	if (!(probFilter & RPMPROB_FILTER_IGNOREARCH) && badArch(rpmteA(p)))
 	    rpmteAddProblem(p, RPMPROB_BADARCH, rpmteA(p), NULL, 0);
@@ -1051,14 +1026,16 @@ static rpmps checkProblems(rpmts ts)
 
 	if (!(probFilter & RPMPROB_FILTER_OLDPACKAGE)) {
 	    Header h;
+	    rpmdbMatchIterator mi;
 	    mi = rpmtsInitIterator(ts, RPMDBI_NAME, rpmteN(p), 0);
 	    while ((h = rpmdbNextIterator(mi)) != NULL)
 		ensureOlder(p, h);
-	    mi = rpmdbFreeIterator(mi);
+	    rpmdbFreeIterator(mi);
 	}
 
 	if (!(probFilter & RPMPROB_FILTER_REPLACEPKG)) {
 	    Header h;
+	    rpmdbMatchIterator mi;
 	    mi = rpmtsInitIterator(ts, RPMDBI_NAME, rpmteN(p), 0);
 	    rpmdbSetIteratorRE(mi, RPMTAG_EPOCH, RPMMIRE_STRCMP, rpmteE(p));
 	    rpmdbSetIteratorRE(mi, RPMTAG_VERSION, RPMMIRE_STRCMP, rpmteV(p));
@@ -1072,10 +1049,13 @@ static rpmps checkProblems(rpmts ts)
 		rpmteAddProblem(p, RPMPROB_PKG_INSTALLED, NULL, NULL,
 				headerGetInstance(h));
 	    }
-	    mi = rpmdbFreeIterator(mi);
+	    rpmdbFreeIterator(mi);
 	}
+
+	if (!(probFilter & RPMPROB_FILTER_FORCERELOCATE))
+	    rpmteAddRelocProblems(p);
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
     return rpmtsProblems(ts);
 }
 
@@ -1087,13 +1067,14 @@ static rpmps checkProblems(rpmts ts)
  */
 static int runTransScripts(rpmts ts, pkgGoal goal) 
 {
+    int rc = 0;
     rpmte p;
     rpmtsi pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, TR_ADDED)) != NULL) {
-	rpmteProcess(p, goal);
+	rc += rpmteProcess(p, goal);
     }
-    pi = rpmtsiFree(pi);
-    return 0; /* what to do about failures? */
+    rpmtsiFree(pi);
+    return rc;
 }
 
 static int rpmtsSetupCollections(rpmts ts)
@@ -1157,7 +1138,7 @@ static int rpmtsSetupCollections(rpmts ts)
 	    }
 	}
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
     /* we've looked at all the rpmte's, update the lastInCollectionAny lists */
     for (i = 0; i < numSeenPost; i++) {
@@ -1213,7 +1194,7 @@ static void addFingerprints(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPr
 	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), rpmfiFC(fi));
 
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
     /* ===============================================
      * Check fingerprints if they contain symlinks
@@ -1233,7 +1214,7 @@ static void addFingerprints(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPr
 	}
 	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), 0);
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
     rpmFpHashFree(symlinks);
 }
@@ -1248,16 +1229,21 @@ static int rpmtsSetup(rpmts ts, rpmprobFilterFlags ignoreSet)
     if (rpmtsFlags(ts) & RPMTRANS_FLAG_NOTRIGGERS)
 	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransTriggers));
 
-    if (rpmtsFlags(ts) & RPMTRANS_FLAG_JUSTDB)
-	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransScripts | _noTransTriggers));
+    if (rpmtsFlags(ts) & (RPMTRANS_FLAG_JUSTDB | RPMTRANS_FLAG_TEST))
+	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransScripts | _noTransTriggers | RPMTRANS_FLAG_NOCOLLECTIONS));
 
-    /* if SELinux isn't enabled, init fails or test run, don't bother... */
+    /* if SELinux isn't enabled or it is a test run, don't bother... */
     if (!is_selinux_enabled() || (rpmtsFlags(ts) & RPMTRANS_FLAG_TEST)) {
         rpmtsSetFlags(ts, (rpmtsFlags(ts) | RPMTRANS_FLAG_NOCONTEXTS));
     }
 
-    if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONTEXTS)) {
-	rpmtsSELabelInit(ts, selinux_file_context_path());
+    if (rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONTEXTS) {
+	rpmlog(RPMLOG_DEBUG, "Selinux disabled.\n");
+    } else {
+	if (rpmtsSELabelInit(ts, 1)) {
+	    rpmlog(RPMLOG_WARNING, "Failed to open SELinux handle.\n");
+	    rpmtsSetFlags(ts, (rpmtsFlags(ts) | RPMTRANS_FLAG_NOCONTEXTS));
+	}
     }
 
     /* 
@@ -1281,7 +1267,7 @@ static int rpmtsSetup(rpmts ts, rpmprobFilterFlags ignoreSet)
 static int rpmtsFinish(rpmts ts)
 {
     if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONTEXTS)) {
-	rpmtsSELabelFini(ts);
+	rpmtsSELabelFini(ts, 1);
     }
     return rpmChrootSet(NULL);
 }
@@ -1294,11 +1280,12 @@ static int rpmtsPrepare(rpmts ts)
     rpmfi fi;
     int rc = 0;
     uint64_t fileCount = countFiles(ts);
+    const char *dbhome = NULL;
+    struct stat dbstat;
 
     fingerPrintCache fpc = fpCacheCreate(fileCount/2 + 10001);
     rpmFpHash ht = rpmFpHashCreate(fileCount/2+1, fpHashFunction, fpEqual,
 			     NULL, NULL);
-    rpmDiskSpaceInfo dsi;
 
     rpmlog(RPMLOG_DEBUG, "computing %" PRIu64 " file fingerprints\n", fileCount);
 
@@ -1313,7 +1300,7 @@ static int rpmtsPrepare(rpmts ts)
 	    skipEraseFiles(ts, p);
 	}
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
 
     /* Open rpmdb & enter chroot for fingerprinting if necessary */
     if (rpmdbOpenAll(ts->rdb) || rpmChrootIn()) {
@@ -1326,7 +1313,10 @@ static int rpmtsPrepare(rpmts ts)
     /* check against files in the rpmdb */
     checkInstalledFiles(ts, fileCount, ht, fpc);
 
-    dsi = rpmtsDbDSI(ts);
+    dbhome = rpmdbHome(rpmtsGetRdb(ts));
+    /* If we can't stat, ignore db growth. Probably not right but... */
+    if (dbhome && stat(dbhome, &dbstat))
+	dbhome = NULL;
 
     pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, 0)) != NULL) {
@@ -1338,31 +1328,41 @@ static int rpmtsPrepare(rpmts ts)
 	   needs on each partition for this package. */
 	handleOverlappedFiles(ts, ht, p, fi);
 
-	rpmtsUpdateDSIrpmDBSize(p, dsi);
-
 	/* Check added package has sufficient space on each partition used. */
 	if (rpmteType(p) == TR_ADDED) {
+	    /*
+	     * Try to estimate space needed for rpmdb growth: guess that the
+	     * db grows 4 times the header size (indexes and all).
+	     */
+	    if (dbhome) {
+		int64_t hsize = rpmteHeaderSize(p) * 4;
+		rpmtsUpdateDSI(ts, dbstat.st_dev, dbhome,
+			       hsize, 0, 0, FA_CREATE);
+	    }
+
 	    rpmtsCheckDSIProblems(ts, p);
 	}
 	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), 0);
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
     rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_STOP, 6, tsmem->orderCount);
 
     /* return from chroot if done earlier */
     if (rpmChrootOut())
 	rc = -1;
 
-    /* File info sets, fp caches etc not needed beyond here, free 'em up. */
-    pi = rpmtsiInit(ts);
-    while ((p = rpmtsiNext(pi, 0)) != NULL) {
-	rpmteSetFI(p, NULL);
+    /* On actual transaction, file info sets are not needed after this */
+    if (!(rpmtsFlags(ts) & (RPMTRANS_FLAG_TEST|RPMTRANS_FLAG_BUILD_PROBS))) {
+	pi = rpmtsiInit(ts);
+	while ((p = rpmtsiNext(pi, 0)) != NULL) {
+	    rpmteSetFI(p, NULL);
+	}
+	rpmtsiFree(pi);
     }
-    pi = rpmtsiFree(pi);
 
 exit:
-    ht = rpmFpHashFree(ht);
-    fpc = fpCacheFree(fpc);
+    rpmFpHashFree(ht);
+    fpCacheFree(fpc);
     rpmtsFreeDSI(ts);
     return rc;
 }
@@ -1382,6 +1382,10 @@ static int rpmtsProcess(rpmts ts)
 	rpmlog(RPMLOG_DEBUG, "========== +++ %s %s-%s 0x%x\n",
 		rpmteNEVR(p), rpmteA(p), rpmteO(p), rpmteColor(p));
 
+	if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONTEXTS)) {
+	    rpmtsSELabelInit(ts, 0);
+	}
+
 	failed = rpmteProcess(p, rpmteType(p));
 	if (failed) {
 	    rpmlog(RPMLOG_ERR, "%s: %s %s\n", rpmteNEVRA(p),
@@ -1390,7 +1394,7 @@ static int rpmtsProcess(rpmts ts)
 	}
 	(void) rpmdbSync(rpmtsGetRdb(ts));
     }
-    pi = rpmtsiFree(pi);
+    rpmtsiFree(pi);
     return rc;
 }
 
@@ -1402,8 +1406,9 @@ int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
     /* Force default 022 umask during transaction for consistent results */
     mode_t oldmask = umask(022);
 
-    /* XXX programmer error segfault avoidance. */
+    /* Empty transaction, nothing to do */
     if (rpmtsNElements(ts) <= 0) {
+	rc = 0;
 	goto exit;
     }
 
@@ -1464,7 +1469,7 @@ exit:
     /* Finish up... */
     (void) umask(oldmask);
     (void) rpmtsFinish(ts);
-    tsprobs = rpmpsFree(tsprobs);
+    rpmpsFree(tsprobs);
     rpmlockFree(lock);
     return rc;
 }
