@@ -1,5 +1,7 @@
 #include "rpmsystem-py.h"
 
+#include <fcntl.h>
+
 #include <rpm/rpmlib.h>	/* rpmReadPackageFile, headerCheck */
 #include <rpm/rpmtag.h>
 #include <rpm/rpmpgp.h>
@@ -15,11 +17,8 @@
 #include "rpmii-py.h"
 #include "rpmps-py.h"
 #include "rpmte-py.h"
-
 #include "rpmts-py.h"
 
-#include "debug.h"
-
 /** \ingroup python
  * \name Class: Rpmts
  * \class Rpmts
@@ -163,7 +162,7 @@ static void die(PyObject *cb)
     if ((r = PyObject_Repr(cb)) != NULL) { 
 	pyfn = PyBytes_AsString(r);
     }
-    fprintf(stderr, _("error: python callback %s failed, aborting!\n"), 
+    fprintf(stderr, "FATAL ERROR: python callback %s failed, aborting!\n", 
 	    	      pyfn ? pyfn : "???");
     rpmdbCheckTerminate(1);
     exit(EXIT_FAILURE);
@@ -279,6 +278,14 @@ rpmts_Clean(rpmtsObject * s)
     Py_RETURN_NONE;
 }
 
+static PyObject *
+rpmts_Clear(rpmtsObject * s)
+{
+    rpmtsEmpty(s->ts);
+
+    Py_RETURN_NONE;
+}
+
 static PyObject *
 rpmts_OpenDB(rpmtsObject * s)
 {
@@ -350,7 +357,7 @@ rpmts_HdrFromFdno(rpmtsObject * s, PyObject *arg)
     	return NULL;
 
     Py_BEGIN_ALLOW_THREADS;
-    rpmrc = rpmReadPackageFile(s->ts, rpmfdGetFd(fdo), "rpmts_HdrFromFdno", &h);
+    rpmrc = rpmReadPackageFile(s->ts, rpmfdGetFd(fdo), NULL, &h);
     Py_END_ALLOW_THREADS;
     Py_XDECREF(fdo);
 
@@ -678,6 +685,9 @@ static struct PyMethodDef rpmts_methods[] = {
   Note: The callback may not be None.\n" },
  {"clean",	(PyCFunction) rpmts_Clean,	METH_NOARGS,
 	NULL },
+ {"clear",	(PyCFunction) rpmts_Clear,	METH_NOARGS,
+"ts.clear() -> None\n\
+Remove all elements from the transaction set\n" },
  {"openDB",	(PyCFunction) rpmts_OpenDB,	METH_NOARGS,
 "ts.openDB() -> None\n\
 - Open the default transaction rpmdb.\n\
@@ -714,7 +724,7 @@ static struct PyMethodDef rpmts_methods[] = {
 - Create a match iterator for the default transaction rpmdb.\n" },
  {"dbIndex",     (PyCFunction) rpmts_index,	METH_VARARGS|METH_KEYWORDS,
 "ts.dbIndex(TagN) -> ii\n\
--Create a key iterator for the default transaction rpmdb.\n" },
+- Create a key iterator for the default transaction rpmdb.\n" },
     {NULL,		NULL}		/* sentinel */
 };
 
@@ -741,7 +751,7 @@ static PyObject * rpmts_new(PyTypeObject * subtype, PyObject *args, PyObject *kw
 
 static int rpmts_init(rpmtsObject *s, PyObject *args, PyObject *kwds)
 {
-    char * rootDir = "/";
+    const char * rootDir = "/";
     rpmVSFlags vsflags = rpmExpandNumeric("%{?__vsflags}");
     char * kwlist[] = {"rootdir", "vsflags", 0};
 
